# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import IOBase, StringIO
from collections import Mapping
from contextlib import contextmanager

from boto3.session import Session

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

try:
    from pathlib import Path
except ImportError:
    import os

    class Path(object):
        def __init__(self, path = '/'):
            self.path = path

        def open(self, *args, **kwargs):
            return open(self.path, *args, **kwargs)

        def __div__(self, path):
            return Path(os.path.join(self.path, path))

@contextmanager
def read(obj):
    """Context manager for reading data from multiple sources as a file object

    Args:
        obj (string|Path|file object): Data to read / read from
                                  If obj is a file object, this is just a pass through
                                  If obj is a Path object, this is similar to obj.open()
                                  If obj is a string, this creates a StringIO so
                                     the data can be read like a file object

    Returns:
        file object: File handle containing data
    """
    try: # Python 2 compatibility
        is_unicode = isinstance(obj, unicode)
    except NameError:
        is_unicode = False

    is_open = False
    if isinstance(obj, Path):
        fh = obj.open()
        is_open = True
    elif isinstance(obj, str) or is_unicode:
        fh = StringIO(obj)
        fh.name = '<string>'
    elif isinstance(obj, IOBase):
        fh = obj
    else:
        raise Exception("Unknown input type {}".format(type(obj).__name__))
    
    try:
        yield fh
    finally:
        if is_open:
            fh.close()

def create_session(**kwargs):
    """Create a Boto3 session from multiple different sources

    Basic file format / dictionary format:
    {
        'aws_secret_key': '',
        'aws_access_key': '',
        'aws_region': '',
        'aws_account_id': ''
    }

    Note: If no arguments are given, a Boto3 session is created and it will attempt
          to figure out this information for itself, from predefined locations.

    Args:
        session (Boto3 Session): Existing session to use. Only account id will be looked for
        credentials (dict|fh|Path|json string): source to load credentials from
                                                If a dict, used directly
                                                If a fh, read and parsed as a Json object
                                                If a Path, opened, read, and parsed as a Json object
                                                If a string, parsed as a Json object

        Note: The following will override the values in credentials if they exist
        secret_key / aws_secret_key (string): AWS Secret Key
        access_key / aws_access_key (string): AWS Access Key

        Note: The following will be derived from the AWS Session if not provided
        account_id / aws_account_id (string): AWS Account ID

        Note: The following will be pulled from EC2 Meta-data if not provided
        region / aws_region (string): AWS region to connect to

    Returns:
        (Boto3 Session, account_id) : Boto3 session populated with given credentials and
                                      AWS Account ID (either given or derived from session)
    """
    credentials = kwargs.get('credentials', {})
    if isinstance(credentials, Mapping):
        pass
    elif isinstance(credentials, Path):
        with credentials.open() as fh:
            credentials = json.load(fh)
    elif isinstance(credentials, str):
        credentials = json.loads(credentials)
    elif isinstance(credentials, IOBase):
        credentials = json.load(credentials)
    else:
        raise Exception("Unknown credentials type: {}".format(type(credentials).__name__))

    def locate(names, locations):
        for location in locations:
            for name in names:
                if name in location:
                    return location[name]
        names = " or ".join(names)
        raise Exception("Could not find credentials value for {}".format(names))

    locate_region = True
    if 'session' in kwargs:
        session = kwargs['session']
        locate_region = False
    else:
        try:
            access = locate(('access_key', 'aws_access_key'), (kwargs, credentials))
            secret = locate(('secret_key', 'aws_secret_key'), (kwargs, credentials))

            session = Session(aws_access_key_id = access,
                              aws_secret_access_key = secret)
        except:
            session = Session() # Let boto3 try to resolve the keys iteself, potentially from EC2 meta data

    try:
        account_id = locate(('account_id', 'aws_account_id'), (kwargs, credentials))
    except:
        # From boss-manage.git/lib/aws.py:get_account_id_from_session()
        account_id = session.client('iam').list_users(MaxItems=1)["Users"][0]["Arn"].split(':')[4]

    if locate_region:
        try:
            region = locate(('region', 'aws_region'), (kwargs, credentials))
        except:
            try:
                region = urlopen('http://169.254.169.254/latest/meta-data/placement/availability-zone/')
                region = region.read().decode('utf-8')[:-1] # remove AZ character from region name
            except:
                raise Exception("Could not locate AWS region to connect to")

        # DP HACK: No good way to get the region after the session is created
        #          Needed here to support loading keys from EC2 meta data
        session._session.set_config_variable('region', region)

    return session, account_id




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class StepFunctionError(Exception):
    """Base class for all library errors"""

    def __init__(self, lineno, pos, msg, line=''):
        super(StepFunctionError, self).__init__(msg)
        self.lineno = lineno
        self.pos = pos
        self.line = line

class TokenError(StepFunctionError):
    @classmethod
    def from_tokenize(cls, error):
        msg, (lineno, pos) = error.args

        return cls(lineno, pos, msg)

class ParserError(StepFunctionError):
    @classmethod
    def from_token(cls, tok, msg):
        lineno, pos = tok.start
        line = tok.line.rstrip() # remove newline
        return cls(lineno, pos, msg, line)

class ActivityError(Exception):
    def __init__(self, error, cause):
        super(ActivityError, self).__init__(cause)
        self.error = error
        self.cause = cause




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tokenize
import token

from .exceptions import TokenError

USELESS = ['NEWLINE', 'NL', 'COMMENT']

class Token(object):
    """Custom class wrapping all token properties"""

    def __init__(self, code, value, start=(0,0), stop=(0,0), line=''):
        """
        Args:
            code (string|int): Token code. Ints are translated using token.tok_name.
            value (string): Token value
            start (tuple): Pair of values describing token start line, start position
            stop (tuple): Pair of values describing token stop line, stop position
            line (string): String containing the line the token was parsed from
        """
        try:
            self.code = token.tok_name[code]
        except:
            self.code = code
        self.value = value
        self.start = start
        self.stop = stop
        self.line = line

    def __str__(self):
        sl, sp = self.start
        el, ep = self.stop
        pos = '{},{}-{},{}'.format(sl, sp, el, ep)
        return '{}, {}, {!r}'.format(pos, self.code, self.value)

    def __repr__(self):
        return '{}({!r}, {!r}, {!r}, {!r}, {!r})'.format(self.__class__.__name__,
                                               self.code,
                                               self.value,
                                               self.start,
                                               self.stop,
                                               self.line)

    def __eq__(self, other):
        return (self.code, self.value) == (other.code, other.value)

def tokenize_source(source):
    """Tokenize a source and convert into Token objects

    Results are filtered by through the list of USELESS tokens

    Args:
        source (callable): Callable object which provides the same interface as io.IOBase.readline
                           Each call provides a line of input as bytes

    Returns:
        list: List of parsed tokens
    """
    try:
        tokens = tokenize.generate_tokens(source)
        tokens = [Token(*t) for t in tokens]
        return [t for t in tokens if t.code not in USELESS]
    except tokenize.TokenError as e:
        raise TokenError.from_tokenize(e)




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from funcparserlib.parser import (some, a, many, skip, maybe, forward_decl)
from funcparserlib.parser import NoParseError, State

from .lexer import Token
from .exceptions import ParserError

from .sfn import Machine
from .sfn import Retry, Catch, Timestamp
from .sfn import PassState, SuccessState, FailState
from .sfn import TaskState, Lambda, Activity
from .sfn import WaitState
from .sfn import ChoiceState, Choice, NotChoice, AndOrChoice
from .sfn import ParallelState, Branch

def link(states, final=None):
    """Take a list of states and links them together in the order given

    Automatically creates a default end state if there is no default and the ChoiceState
    terminates.

    States for ChoiceState branches (and catch blocks) are processed and added to the list
    of returned states.

    Convention: If the State has an attribute 'branches' is should be a list of list of state.
                Each list of state will be linked together and then appended to the list of
                returned States.

    Args:
        states (list): List of States to link together
        final (string|State|None): Name of next state to link the final state to
                                   None to terminate execution with the final state

    Returns:
        list: List of linked States
    """
    linked = []
    for i in range(len(states)):
        state = states[i]

        # DP TODO: check to see if the given state name already exists
        linked.append(state)

        next_ = states[i+1] if i+1 < len(states) else final

        # The first three conditions are checking to see if the state needs
        # to be linked with the next state or if it is already linked / terminates
        if 'Next' in state:
            pass # process branches, if they exist
        elif 'End' in state:
            pass # process branches, if they exist
        elif type(state) in (SuccessState, FailState): #terminal states
            pass
        elif type(state) == ChoiceState:
            if 'Default' not in state:
                next__ = next_ # prevent branches from using the new end state (just use End=True)
                if next__ is None:
                    next__ = SuccessState(str(state) + "Next")
                    linked.append(next__)
                state['Default'] = str(next__)
        else:
            if next_ is None:
                state['End'] = True
            else:
                state['Next'] = str(next_)

        if hasattr(state, 'branches'):
            for branch in state.branches:
                linked.extend(link(branch, final=next_))

    return linked

def make_name(line):
    """Take the line number of a state and return the state name"""
    return "Line{}".format(line)

def add_name_comment(state, comment):
    """Take a doc string comment and set the state name and comment

    Args:
        state (State): State to update the name of and add the comment to
        comment (None|tuple): None or tuple of (name, comment)
                              If name is an empty string no name change is performed
    """
    if comment:
        name, comment = comment.split('\n', 1)

        name = name.strip()
        # DP TODO: Remove indent from each line of the comment, preserving internal indent
        comment = '\n'.join([l.strip() for l in comment.split('\n')])

        if  len(name) > 0:
            if type(state) == ChoiceState:
                # update while loops to use the new name
                if state.branches[0][-1]['Next'] == state.Name:
                    state.branches[0][-1]['Next'] = name
            state.Name = name

        state['Comment'] = comment

# Dictionary of comparison operator and type that resolves
# to the state machine comparison operator
COMPARISON = {
    '==': {
        str: 'StringEquals',
        int: 'NumericEquals',
        float: 'NumericEquals',
        bool: 'BooleanEquals',
        Timestamp: 'TimestampEquals',
    },
    '<': {
        str: 'StringLessThan',
        int: 'NumericLessThan',
        float: 'NumericLessThan',
        Timestamp: 'TimestampLessThan',
    },
    '>': {
        str: 'StringGreaterThan',
        int: 'NumericGreaterThan',
        float: 'NumericGreaterThan',
        Timestamp: 'TimestampGreaterThan',
    },
    '<=': {
        str: 'StringLessThanEquals',
        int: 'NumericLessThanEquals',
        float: 'NumericLessThanEquals',
        Timestamp: 'TimestampLessThanEquals',
    },
    '>=': {
        str: 'StringGreaterThanEquals',
        int: 'NumericGreaterThanEquals',
        float: 'NumericGreaterThanEquals',
        '': 'TimestampGreaterThanEquals',
    },
}

const = lambda x: lambda _: x
tokval = lambda x: x.value
tokline = lambda x: x.start[0]
toklineval = lambda x: (x.start[0], x.value)
toktype = lambda t: some(lambda x: x.code == t) >> tokval
op = lambda s: a(Token('OP', s)) >> tokval
op_ = lambda s: skip(op(s))
n = lambda s: a(Token('NAME', s)) >> tokval
n_ = lambda s: skip(n(s))
l = lambda s: a(Token('NAME', s)) >> tokline
# extract both line number and token value at the same time
ln = lambda s: a(Token('NAME', s)) >> toklineval

end = skip(a(Token('ENDMARKER', '')))
block_s = skip(toktype('INDENT'))
block_e = skip(toktype('DEDENT'))

def debug(x):
    """Print the current object being parsed"""
    print(x)
    return x

def debug_(m):
    """Print the current object with a prefix

    Args:
        m (string): Prefix to print before debuged object
    """
    def debug__(a):
        print("{}: {!r}".format(m, a))
        return a
    return debug__

def make_number(n):
    """Parse the given string as an int or float"""
    try:
        return int(n)
    except ValueError:
        return float(n)

def make_string(n):
    """Remove a strings quotes and return the raw string

    NOTE: escaping is done / undone
    """
    if n[:3] in ('"""', "'''"):
        s = n[3:-3]
    else:
        s = n[1:-1]
    return s

def make_ts_str(s):
    """Check to see if the given string is a Timestamp

    Returns:
        Timestamp|string: Timestamp if the string is a valid timestamp
                          String if the the string is not a timestamp
    """
    try:
        # A timestamp is also a valid string, so it has to be manually parsed
        # instead of using the lexer
        return Timestamp(s)
    except:
        return s

def make_array(n):
    """Take the results of parsing an array and return an array

    Args:
        n (None|list): None for empty list
                       list should be [head, [tail]]
    """
    if n is None:
        return []
    else:
        return [n[0]] + n[1]

def make_object(n):
    """Take a list of pairs and create a dict
    
    NOTE: run through make_array to transform the results to an array
    """
    return dict(make_array(n))

# =============
# Simple States
# =============
def make_pass(args):
    """Make a PassState

    Args:
        args (int): Line number
    """
    line = args

    name = make_name(line)
    state = PassState(name)
    state.line = line
    return state

def make_success(args):
    """Make a SuccessState

    Args:
        args (int): Line number
    """
    line = args

    name = make_name(line)
    state = SuccessState(name)
    state.line = line
    return state

def make_fail(args):
    """Make a FailState

    Args:
        args (tuple): (line number:int, error:string, cause:string)
    """
    line, error, cause = args

    name = make_name(line)
    state = FailState(name, error, cause)
    state.line = line
    return state

# make_task moved into parse function to have access to parse arguments

def make_wait(args):
    """Make a WaitState

    Args:
        args (tuple): (line number:int, key:string, value:string|int|Timestamp)
    """
    line, key, value = args

    if key == 'timestamp' and type(value) != Timestamp:
        raise ParserError(line, 0, "Invalid timestamp '{}'".format(value), '')

    name = make_name(line)
    kwargs = {key: value}

    state = WaitState(name, **kwargs)
    state.line = line
    return state

# ============
# Flow Control
# ============
def make_comp_simple(args):
    """Make a Choice statement for a ChoiceState

    Args:
        args (tuple): (var:string, op:string, val:string|int|Timestamp)
    """
    var, op, val = args
    
    if op == '!=':
        op = COMPARISON['=='][type(val)]
        choice = Choice(var, op, val)
        return NotChoice(choice)
    else:
        op = COMPARISON[op][type(val)]
        return Choice(var, op, val)

def make_comp_not(args):
    """Wrap the given Choice in a NotChoice"""
    return NotChoice(args)

def make_comp_andor(args):
    """Process a list of and / or combinding Choices into an AndOrChoice

    Args:
        args (tuple): (Choice, [('and'|'or', Choice)])
                      The 'and' or 'or' operator should be
                      the same for all tuples in the list
    """
    x, xs = args

    if len(xs) == 0:
        return x

    vals = [x]
    op = ''
    for op_, val in xs:
        op = op_
        vals.append(val)

    return AndOrChoice(op.capitalize(), vals)

def make_while(args):
    """Make a while loop into a ChoiceState

    Args
        args (tuple): (line:int, choice:Choice, (comment:string, steps:[State]))
                      choice: The loop conditional
                      steps: The loop body
    """
    line, choice, (comment, steps) = args
    name = make_name(line)

    choice['Next'] = str(steps[0])
    choices = ChoiceState(name, [choice])
    choices.line = line
    choices.branches = [steps]
    steps[-1]['Next'] = name # Create the loop
    add_name_comment(choices, comment)
    return choices

def make_if_else(args):
    """Make an if/elif/else into a ChoiceState

    Args
        args (tuple): (line:int, choice:Choice, (comment:string, steps:[State]), elif, else)
                      choice: The loop conditional
                      steps: The loop body
                      elif [(line:int, choice:Choice, (comment:string, steps:[State]))] The elif blocks (same as the if block)
                      else (line:int, (comment:string, steps:[State])) The else block
    """
    line, choice, (comment, steps), elif_, else_ = args

    branches = []
    choices = []

    choice['Next'] = str(steps[0])
    branches.append(steps)
    choices.append(choice)

    for line_, choice_, (_, steps_) in elif_:
        choice_['Next'] = str(steps_[0])
        branches.append(steps_)
        choices.append(choice_)

    if else_:
        line_, (_, steps_) = else_
        branches.append(steps_)
        default = str(steps_[0])
    else:
        default = None

    name = make_name(line)
    state = ChoiceState(name, choices, default)
    state.line = line
    state.branches = branches
    add_name_comment(state, comment)
    return state

def make_switch(args):
    """Make a switch statement into a ChoiceState

    Args
        args (tuple): (line:int,
                       var:string,
                       comment:string,
                       (val:int|float|string|ts, (comment:string, steps:[State])),
                       (comment:string, steps:[State]))

                      var: JsonPath of the variable to use in the case comparisons
                      val: Value to compare the variable against
                      steps: The case / default body
    """
    line, var, comment, cases, default = args

    choices = []
    branches = []

    for val, (_, steps_) in cases:
        choice = make_comp_simple((var, '==', val))
        choice['Next'] = str(steps_[0])
        branches.append(steps_)
        choices.append(choice)

    if default:
        (_, steps_) = default
        branches.append(steps_)
        default = str(steps_[0])

    name = make_name(line)
    state = ChoiceState(name, choices, default)
    state.line = line
    state.branches = branches
    add_name_comment(state, comment)
    return state

def make_parallel(args):
    """Make a ParallelState

    Args:
        args (tuple): (line:int, (comment:string, steps:[State]), parallels)
                      steps: The parallel body
                      parallels [(line:int, (comment:string, steps:[State]))] Additional parallel blocks
    """
    line, (comment, steps), parallels = args

    branches = []

    #DP ???: calling link in the middle of parsing. should this be called after all states are parsed??
    branches.append(Branch(link(steps), str(steps[0])))

    for line_, (_, steps_) in parallels:
        branches.append(Branch(link(steps_), str(steps_[0])))

    name = make_name(line)
    state = ParallelState(name, branches)
    state.line = line
    add_name_comment(state, comment)
    return state

def make_retry(args):
    """Make a Retry statement

    Args:
        args (tuple): (errors:string|list, interval:int, max:int, backoff:int|float)
                      errors: Single error or list of errors to catch
                              If an empty list, replaced with ['States.ALL']
                      interval: retry interval (seconds) for first retry
                      max: total number of retry attempts
                      backoff: multiplier applied to interval for each retry attempt
    """
    errors, interval, max_, backoff = args

    if errors == []:
        errors = ['States.ALL'] # match all errors if none is given
    return Retry(errors, interval, max_, backoff)

def make_catch(args):
    """Make a Catch statement

    Args:
        args (tuple): (errors:string|list, results:None|string, steps:[State])
                      errors: Single error or list of errors to catch
                              If an empty list, replaced with ['States.ALL']
                      results: ResultPath for where to place error information
                               in the input for the next state
                      steps: The catch body
    """
    errors, result_path, steps = args

    next_ = str(steps[0])

    if errors == []:
        errors = ['States.ALL'] # match all errors if none is given
    catch = Catch(errors, next_, result_path)
    catch.branches = [steps]
    return catch

def make_modifiers(args):
    """Take a mixed list of Catch and Retry statements and return a tuple of lists

    Args:
        args (list): List of Retry and Catch objects

    Returns:
        (retry, catch): retry is None or a list of Retry objects
                        catch is None or a list of Catch objects
    """
    retry = []
    catch = []
    for modifier in args:
        if type(modifier) == Retry:
            retry.append(modifier)
        elif type(modifier) == Catch:
            catch.append(modifier)
        else:
            raise Exception("Unknown modifier type: " + type(modifier).__name__)
    if retry == []:
        retry = None
    if catch == []:
        catch = None
    return (retry, catch)

def make_flow_modifiers(args):
    """Prepare flow control statements to be passed to add_modifiers

    Args:
        args (tuple): (state:State, transform:, errors:)
                      state: State to add the transform and error modifiers to
                      transform:
                      errors: An optional block
    Returns:
        tuple: Tuple formated for input to add_modifiers()
    """
    try:
        state, transform, errors = args
    except:
        state, transform = args
        errors = None

    return (state, (None, None, None, transform, None, errors))

def add_modifiers(args):
    """Add modifiers to a state

    Note all modifiers can be applied to all states. If a modifier is given
    that cannot be applied to the given state, and exception is raised.

    Args:
        args (tuple): (state:State,
                       (
                            comment:string,
                            timeout:int
                            heartbeat:int,
                            (
                                input:string,
                                 result:string,
                                 output:string
                            ),
                            data:dict,
                            (
                                retries:[Retry],
                                catches:[Catch]
                            )
                       )
                      )
                      timeout: task timeout
                      heartbeat: task heartbeat
                      input: input JsonPath selector
                      result: result JsonPath selector
                      output: output JsonPath selector
                      data: PassState Json results
                      retries: list of Retries
                      catches: list of Catches
    """
    state, args = args

    type_ = type(state)

    # DP TODO: Create AST so that ParserError context can be filled in
    error = lambda m: ParserError(state.line, 0, m, '')

    if args:
        comment, timeout, heartbeat, transform, data, modifiers = args

        add_name_comment(state, comment)

        if timeout:
            if type_ not in (TaskState,):
                raise error("Invalid modifier 'timeout'")
            state['TmeoutSeconds'] = timeout
        else:
            timeout = 60

        if heartbeat:
            if type_ not in (TaskState,):
                raise error("Invalid modifier 'heartbeat'")
            if not heartbeat < timeout:
                raise error("Modifier 'heartbeat' must be less than 'timeout'")
            state['HeartbeatSeconds'] = heartbeat

        if transform:
            input_path, result_path, output_path = transform

            if input_path:
                if type_ in (FailState,):
                    raise error("Invalid modifier 'input'")
                state['InputPath'] = input_path

            if result_path:
                if type_ in (FailState, SuccessState, WaitState):
                    raise error("Invalid modifier 'result'")
                state['ResultPath'] = result_path

            if output_path:
                if type_ in (FailState,):
                    raise error("Invalid modifier 'output'")
                state['OutputPath'] = output_path

        if data:
            if type_ != PassState:
                raise error("Invalid modifier 'data'")
            state['Result'] = data

        if modifiers:
            retries, catches = modifiers
            if retries:
                if type_ not in (TaskState, ParallelState):
                    raise error("Invalid modifier 'retry'")
                state['Retry'] = retries
            if catches:
                if type_ not in (TaskState, ParallelState):
                    raise error("Invalid modifier 'catch'")
                state['Catch'] = catches
                state.branches = []
                for catch in catches:
                    state.branches.extend(catch.branches)

    return state

def json_text():
    """Returns the parser for Json formatted data"""
    # Taken from https://github.com/vlasovskikh/funcparserlib/blob/master/funcparserlib/tests/json.py
    # and modified slightly
    null = (n('null') | n('Null')) >> const(None)
    true = (n('true') | n('True')) >> const(True)
    false = (n('false') | n('False')) >> const(False)
    number = toktype('NUMBER') >> make_number
    string = toktype('STRING') >> make_string

    value = forward_decl()
    member = string + op_(u':') + value >> tuple
    object = (
        op_(u'{') +
        maybe(member + many(op_(u',') + member) + maybe(op_(','))) +
        op_(u'}')
        >> make_object)
    array = (
        op_(u'[') +
        maybe(value + many(op_(u',') + value) + maybe(op_(','))) +
        op_(u']')
        >> make_array)

    value.define(
        null
        | true
        | false
        | object
        | array
        | number
        | string)
    json_text = object | array

    return json_text

def parse(seq, region=None, account=None, translate=lambda x: x):
    """Parse the given sequence of tokens into a StateMachine object

    Args:
        seq (list): List of lexer.Token tokens to parse
        region (string): AWS region for constructed Lambda/Activity ARNs
        account (string): AWS account id for constructed Lambda/Activity ARNs
        translate (function): Translation function applied to Lambda/Activity names
                              before ARNs are constructed

    Returns
        sfn.StateMachine: StateMachine object
    """
    def make_task(args):
        """Make a TaskState

        Args:
            args (tuple): ((line number:int, type:string), function:string)
                          type: 'Lambda' or 'Activity'
                          function: Full or partial ARN of the Lambda or Activity
        """
        (line, type_), func = args

        func = translate(func)

        name = make_name(line)
        try:
            if type_ == "Lambda":
                task = Lambda(func, region, account)
            elif type_ == "Activity":
                task = Activity(func, region, account)
            else:
                raise Exception("Unsuported task type '{}'".format(type_))
        except Exception as e:
            raise ParserError(line, 0, str(e), '')

        state = TaskState(name, task)
        state.line = line
        return state

    state = forward_decl()

    # Primatives
    number = toktype('NUMBER') >> make_number
    string = toktype('STRING') >> make_string
    ts_str = toktype('STRING') >> make_string >> make_ts_str
    array = op_('[') + maybe(string + many(op_(',') + string)) + op_(']') >> make_array
    retry = n_('retry') + (array|string) + number + number + number >> make_retry
    catch = n_('catch') + (array|string) + op_(':') + maybe(string) + block_s + many(state) + block_e >> make_catch

    # Transform / Error statements
    path = lambda t: maybe(n_(t) + op_(':') + string)
    mod_transform = path('input') + path('result') + path('output')
    data = n_('data') + op_(':') + block_s + json_text() + block_e
    modifier = retry | catch
    mod_error = maybe(modifier + many(modifier) >> make_array >> make_modifiers)
    modifiers = (block_s +
                    maybe(string) +
                    maybe(n_('timeout') + op_(':') + number) + 
                    maybe(n_('heartbeat') + op_(':') + number) + 
                    mod_transform +
                    maybe(data) +
                    mod_error +
                 block_e)

    # Simple States
    pass_ = l('Pass') + op_('(') + op_(')') >> make_pass
    success = l('Success') + op_('(') + op_(')') >> make_success
    fail = l('Fail') + op_('(') + string + op_(',') + string + op_(')') >> make_fail
    task = (ln('Lambda') | ln('Activity')) + op_('(') + string + op_(')') >> make_task
    wait_types = n('seconds') | n('seconds_path') | n('timestamp') | n('timestamp_path')
    wait = l('Wait') + op_('(') + wait_types + op_('=') + (number|ts_str) + op_(')') >> make_wait
    simple_state = pass_ | success | fail | task | wait
    simple_state_ = simple_state + maybe(modifiers) >> add_modifiers

    # Flow Control blocks
    transform_block = maybe(n_('transform') + op_(':') + block_s + maybe(mod_transform) + block_e)
    error_block = maybe(n_('error') + op_(':') + block_s + maybe(mod_error) + block_e)
    block = block_s + maybe(string) + many(state) + block_e

    # Comparison logic
    comp_op = op('==') | op('<') | op('>') | op('<=') | op('>=') | op('!=')
    comp_simple = string + comp_op + (number|ts_str) >> make_comp_simple

    comp_stmt = forward_decl()
    comp_base = forward_decl()
    comp_base.define((op_('(') + comp_stmt + op_(')')) | comp_simple | ((n_('not') + comp_base) >> make_comp_not))
    comp_and = comp_base + many(n('and') + comp_base) >>  make_comp_andor
    comp_or = comp_and + many(n('or') + comp_and) >> make_comp_andor
    comp_stmt.define(comp_or)

    # Control Flow states
    comparison = comp_stmt + op_(':')
    while_ = l('while') + comparison + block >> make_while
    if_else = (l('if') + comparison + block +
               many(l('elif') + comparison + block) +
               maybe(l('else') + op_(':') + block)) >> make_if_else
    case = n_('case') + (number|ts_str) + op_(':') + block
    switch = (l('switch') + string + op_(':') +
              block_s + maybe(string) + many(case) +
              maybe(n_('default') + op_(':') + block) +
              block_e) >> make_switch
    choice_state = while_ | if_else | switch
    choice_state_ = (choice_state + transform_block) >> make_flow_modifiers >> add_modifiers
    parallel = (l('parallel') + op_(':') + block +
                many(l('parallel') + op_(':') + block)) >> make_parallel
    parallel_ = (parallel + transform_block + error_block) >> make_flow_modifiers >> add_modifiers
    state.define(simple_state_ | choice_state_ | parallel_)

    # State Machine
    version = maybe(n_('version') + op_(':') + string)
    timeout = maybe(n_('timeout') + op_(':') + number)
    machine = maybe(string) + version + timeout + many(state) + end

    try:
        # call run() directly to have better control of error handling
        (tree, _) = machine.run(seq, State())
        comment, version_, timeout_, states_ = tree
        states_ = link(states_)

        return Machine(comment = comment,
                       states = states_,
                       start = states_[0],
                       version = version_,
                       timeout = timeout_)
    except NoParseError as e:
        max = e.state.max
        tok = seq[max] if len(seq) > max else Token('EOF', '<EOF>')

        if tok.code == 'ERRORTOKEN':
            msg = "Unterminated quote"
        else:
            msg = "Invalid syntax"
            # DP ???: Should the actual token be used in the error message?

        raise ParserError.from_token(tok, msg)




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
import time
import random
import json
from string import ascii_uppercase as CHARS
from multiprocessing import Process

from botocore.exceptions import ClientError

from .exceptions import ActivityError
from .utils import create_session

# DP TODO: rewrite as Mixins so that Processes, Threads, and other implementations can be used

class Activity(object):
    """Class for work with and being an AWS Step Function Activity"""

    def __init__(self, name, arn=None, worker=None, **kwargs):
        """
        Args:
            name (string): Name of the Activity
            arn (string): ARN of the Activity (None to have it looked up)
            worker (string): Name of the worker receiving tasks (None to have one created)
            kwargs (dict): Same arguments as utils.create_session()
        """
        self.name = name
        self.arn = arn
        self.worker = worker or (name + "-" + "".join(random.sample(CHARS, 6)))
        self.token = None
        self.session, self.account_id = create_session(**kwargs)
        self.client = self.session.client('stepfunctions')

        if self.arn is None:
            resp = self.client.list_activities()
            for activity in resp['activities']:
                if activity['name'] == name:
                    self.arn = activity['activityArn']
                    break
        else:
            try:
                resp = self.client.describe_activity(activityArn = self.arn)
                if resp['name'] != name:
                    raise Exception("Name of {} is not {}".format(self.arn, self.name))
            except ClientError:
                raise Exception("ARN {} is not valid".format(self.arn))

    @property
    def exists(self):
        """If the Activity exist (has an ARN in AWS)"""
        return self.arn is not None

    def create(self, exception = False):
        """Create the Activity in AWS

        Args:
            exception (boolean): If an exception should be raised if the Activity already exists (default: False)
        """
        if self.exists:
            if exception:
                raise Exception("Activity {} already exists".format(self.name))
        else:
            resp = self.client.create_activity(name = self.name)
            self.arn = resp['activityArn']

    def delete(self, exception = False):
        """Delete the Activity from AWS

        Args:
            exception (boolean): If an exception should be raised if the Activity doesn't exists (default: False)
        """
        if not self.exists:
            if exception:
                raise Exception("Activity {} doesn't exist".format(self.name))

        resp = self.client.delete_activity(activityArn = self.arn)
        self.arn = None

    def task(self):
        """Query to see if a task exists for processing.

        This methods uses a long poll, waiting up to 60 seconds before returning

        Returns:
            dict|None: Json dictionary of arguments or None if there is no task yet
        """
        if self.token is not None:
            raise Exception("Currently working on a task")

        resp = self.client.get_activity_task(activityArn = self.arn,
                                             workerName = self.worker)

        if len(resp['taskToken']) == 0:
            return None
        else:
            self.token = resp['taskToken']
            return json.loads(resp['input'])

    def success(self, output):
        """Marks the task successfully complete and returns the processed data

        Args:
            output (string|dict): Json response to return to the state machine
        """
        if self.token is None:
            raise Exception("Not currently working on a task")

        output = json.dumps(output)

        resp = self.client.send_task_success(taskToken = self.token,
                                             output = output)
        self.token = None # finished with task

    def failure(self, error, cause):
        """Marks the task as a failure with a given reason

        Args:
            error (string): Failure error
            cause (string): Failure error cause
        """
        if self.token is None:
            raise Exception("Not currently working on a task")

        resp = self.client.send_task_failure(taskToken = self.token,
                                             error = error,
                                             cause = cause)
        self.token = None # finished with task

    def heartbeat(self):
        """Sends a heartbeat for states that require heartbeats of long running Activities"""
        if self.token is None:
            raise Exception("Not currently working on a task")

        resp = self.client.send_task_heartbeat(taskToken = self.token)

class TaskProcess(Process):
    """Process for processing an Activity's task input"""

    def __init__(self, worker, token, input_, target=None, **kwargs):
        """
        Args:
            worker (string): Name of the activity work that accepted the tasking
            token (string): AWS StepFunction token associated with the accepted tasking
            input_ (Json): Json input to the activity invocation
            target (callable|None): Function to call to process the input_ data
                                    If target returns a generator corouting a heartbeat
                                    is sent each time the function yields control
            kwargs (dict): Same arguments as utils.create_session()
        """
        super(TaskProcess, self).__init__(name=worker)
        self.worker = worker
        self.token = token
        self.input_ = input_
        self.target = target
        self.session, self.account_id = create_session(**kwargs)
        self.client = self.session.client('stepfunctions')

    def process(self, input_):
        """Call the target function or raise an exception.

        Designed to be overriden by a child class to provide an easy hook for processing
        data.

        Args:
            input_ (Json): Input Json data passed to the constructor

        Returns:
            Generator Coroutine : This activity wants a heartbeat to be sent each time
                                  the coroutine yields control
            object : The results of processing the input data, to be passed to success()
        """
        if self.target:
            return self.target(input_)
        else:
            # DP ???: Use ActivityError('Unhandled', ...)???
            raise Exception("No target to handle processing")

    def run(self):
        """Called by Process.start() to process data

        Note: If the task times out, the task results (or running coroutine)
              are silently discarded.
        """
        try:
            output_ = self.process(self.input_)
            # DP TODO: Add coroutine support
            if isinstance(output_, types.GeneratorType):
                try:
                    it = output_
                    while True:
                        output_ = next(it)
                        self.heartbeat()
                except StopIteration as e:
                    # Support getting the output from both Python 2 & 3 generators
                    if hasattr(e, 'value'):
                        output_ = e.value
            self.success(output_)
        except ActivityError as e:
            self.failure(e.error, e.cause)
        except Exception as e:
            if self.is_timeout(e): # ClientError
                self.token = None
                return # Eat timeout error from heartbeat

            error = type(e).__name__
            self.failure(error, str(e))

    def is_timeout(self, ex, op_name=None):
        """Check the exception to determine if it is a Boto3 ClientError
        thrown because the task timed out.

        Args:
            ex (Exception) : Exception to check
            op_name (string|None) : (Optional) name of the operation that was attempted

        Returns:
            bool : If the exception was caused by the task timing out
        """
        try:
            rst = ex.response['Error']['Code'] == 'TaskTimedOut'
            if op_name:
                rst &= ex.operation_name == op_name
            return rst
        except:
            return False

    def success(self, output):
        """Marks the task successfully complete and returns the processed data

        Note: This method will silently fail if the task has timed out

        Args:
            output (string|dict): Json response to return to the state machine
        """
        if self.token is None:
            raise Exception("Not currently working on a task")

        output = json.dumps(output)

        try:
            resp = self.client.send_task_success(taskToken = self.token,
                                                 output = output)
        except ClientError as e:
            # eat the timeout
            if not self.is_timeout(e):
                raise
        finally:
            self.token = None # finished with task

    def failure(self, error, cause):
        """Marks the task as a failure with a given reason

        Note: This method will silently fail if the task has timed out

        Args:
            error (string): Failure error
            cause (string): Failure error cause
        """
        if self.token is None:
            raise Exception("Not currently working on a task")

        try:
            resp = self.client.send_task_failure(taskToken = self.token,
                                                 error = error,
                                                 cause = cause)
        except ClientError as e:
            # eat the timeout
            if not self.is_timeout(e):
                raise
        finally:
            self.token = None # finished with task

    def heartbeat(self):
        """Sends a heartbeat for states that require heartbeats of long running Activities"""
        if self.token is None:
            raise Exception("Not currently working on a task")

        resp = self.client.send_task_heartbeat(taskToken = self.token)
        # heartbeat error handled in the run() method

class ActivityProcess(Process):
    """Process for polling an Activity's ARN and launching TaskProcesses to handle
    the given tasking
    """

    def __init__(self, name, task_proc, **kwargs):
        """
        Note: task_proc arguments are the same as TaskProcess (minus the target argument)
        Note: task_proc must return an object with a start() method

        Args:
            name (string): Name of the activity to monitor
                           The activity's ARN is looked up in AWS using the provided
                           AWS credentials
            task_proc (string|callable): Callable that produces a Process to run
                                         If string, the class / function will be imported
            kwargs (dict): Same arguments as utils.create_session()
        """
        super(ActivityProcess, self).__init__(name=name)
        self.name = name
        self.credentials = kwargs
        self.session, self.account_id = create_session(**kwargs)
        self.client = self.session.client('stepfunctions')
        self.arn = None

        resp = self.client.list_activities()
        for activity in resp['activities']:
            if activity['name'] == name:
                self.arn = activity['activityArn']
                break
        
        if isinstance(task_proc, str):
            import importlib
            module, _, function = task_proc.rpartition('.')
            module = importlib.import_module(module)
            task_proc = module.__dict__[function]
        self.task_proc = task_proc

    def run(self):
        """Called by Process.start() to process data

        Note: Automatically creates the Activity ARN if it doesn't exist

        """
        self.create()

        get_worker = lambda: (self.name + '-' + ''.join(random.sample(CHARS,6)))

        worker = get_worker()

        # Note: needed for unit test, so the loop will exit
        self.running = True
        while self.running:
            try:
                token, input_ = self.task(worker)
                if token is not None:
                    proc = self.task_proc(worker, token, input_, **self.credentials)
                    proc.start()
                    # DP TODO: figure out how to limit the total number of currently running task processes
                    # DP ???: create a list of launched processes, so on terminate the handling tasks will terminate too
                    #         if so, need to figure out how to send a failure for the tasks that havn't finished...

                    worker = get_worker()
            except Exception as e:
                # DP ???: create a new worker name?
                # DP ???: create a flag for if a task was accepted and fail it if there was an issue launching the task
                # DP ???: What to do when there is an exception communicating with AWS
                #         Stop running, wait, just loop and continue to fail?
                pass # DP TODO: figure out logging to syslog

    def create(self, exception = False):
        """Create the Activity in AWS

        Args:
            exception (boolean): If an exception should be raised if the Activity already exists (default: False)
        """
        if self.arn is not None:
            if exception:
                raise Exception("Activity {} already exists".format(self.name))
        else:
            resp = self.client.create_activity(name = self.name)
            self.arn = resp['activityArn']

    def delete(self, exception = False):
        """Delete the Activity from AWS

        Args:
            exception (boolean): If an exception should be raised if the Activity doesn't exists (default: False)
        """
        if self.arn is None:
            if exception:
                raise Exception("Activity {} doesn't exist".format(self.name))
        else:
            resp = self.client.delete_activity(activityArn = self.arn)
            self.arn = None

    def task(self, worker):
        """Query to see if a task exists for processing.

        This methods uses a long poll, waiting up to 60 seconds before returning

        Args:
            worker (string): Name of the worker process to receive tasking for

        Returns:
            string|None, dict|None: Task token, Json dictionary of arguments
                                    Nones if there is no task yet
        """
        if self.arn is None:
            raise Exception("Activity {} doesn't exist".format(self.name))

        resp = self.client.get_activity_task(activityArn = self.arn,
                                             workerName = worker)

        if len(resp['taskToken']) == 0:
            return None, None
        else:
            return resp['taskToken'], json.loads(resp['input'])

class ActivityManager(object):
    """Manager for launching multiple ActivityProcesses and monitoring that
    they are still running.

    Designed to be subclasses and customized to provide a list of processes to start
    and monitor.
    """

    def __init__(self):
        self.is_running = False

    def build(self):
        """Build a list of callables that return a process to monitor

        Each callable is called once to create the inital process and
        then each time the current process dies.

        Should be overridden by the subclass.

        Returns:
            List of callables
        """
        return []

    def run(self):
        """Start the initial set of processes and monitor them to ensure
        they are still running. If any have stopped, the method will restart
        the processes.

        Note: This is a blocking method. It will continue to run until
        self.is_running is False
        """
        activities = self.build()
        if len(activities) == 0:
            raise Exception("No activities to manage")

        self.is_running = True

        procs = {}
        try:
            for a in activities:
                procs[a] = a()
                procs[a].start()

            while self.is_running:
                for key in procs:
                    if not procs[key].is_alive():
                        procs[key] = key()
                        procs[key].start()
                time.sleep(60)
        except KeyboardInterrupt:
            for key in procs:
                if procs[key].is_alive():
                    procs[key].terminate()




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import OrderedDict

import iso8601 # parser for timestamp format

class _StateMachineEncoder(json.JSONEncoder):
    """Custom JSONEncoder that handles the Timestamp type"""
    def default(self, o):
        if type(o) == Timestamp:
            return str(o)
        return super(_StateMachineEncoder, self).default(o)

class Branch(dict):
    """A branch of execution. A list of self contains states (only references 
    to states within the list) and a pointer to the first state to start
    execution on
    """

    def __init__(self, states=None, start=None):
        """
        Args:
            states (None|list|dict): No States, a list of States, or dict of
                                     States for the branch to execute
            start (None|State|string): The first State to start execution on
        """
        super(Branch, self).__init__()
        # Makes states be dumped in the same order they were added
        # making it easier to read the output and match it to the input
        self['States'] = OrderedDict() 

        if type(states) == list:
            for state in states:
                self.addState(state)
        elif type(states) == dict:
            self['States'].update(states)

        if start is not None:
            self.setStart(start)

    def setStart(self, state):
        """Set the start State

        Args:
            state (State|string): State / Name of State to start execution on
        """
        self['StartAt'] = str(state)

    def addState(self, state):
        """Add a State to the list of States available to execute

        Args:
            state (State): State to add
        """
        self['States'][state.Name] = state


class Machine(Branch):
    """State Machine, consisting of a single branch of execution"""

    def __init__(self, comment=None, states=None, start=None, version=None, timeout=None):
        """
        Args:
            comment (string): State machine comment
            state (list): List of States to execute
            start (State|string): State to start execution on
            version (string): AWS State MAchine Language version used
            timeout (int): Overall state machine timeout
        """
        super(Machine, self).__init__(states, start)

        if comment is not None:
            self['Comment'] = comment

        if version is not None:
            self['Version'] = version

        if timeout is not None:
            self['TimeoutSeconds'] = timeout

    def definition(self, **kwargs):
        """Dump the state machine into the JSON format needed by AWS

        Args:
            kwargs (dict): Arguments passed to json.dumps()
        """
        return json.dumps(self, cls=_StateMachineEncoder, **kwargs)

class State(dict):
    """Base class for all States"""

    def __init__(self, name, type_, next=None, end=None, catches=None, retries=None):
        """
        Args:
            name (string): Name of the state
            type_ (string): Type of the State
            next (string|State): Next state to transition to
            end = (boolean): If this state is a terminal state
            catches (Catch|[Catch]): Catch(s) for the State
            retires (Retry|[Retry]): Retry(s) for the State
        """
        super(State, self).__init__(Type = type_)
        self.Name = name

        if next is not None:
            self['Next'] = str(next)

        if end is not None:
            self['End'] = bool(end)

        if catches is not None:
            if type(catches) != list:
                catches = [catches]
            self['Catch'] = catches

        if retries is not None:
            if type(retries) != list:
                retries = [retries]
            self['Retry'] = retries

    def addCatch(self, catch):
        """Add a Catch to this State"""
        if 'Catch' not in self:
            self['Catch'] = []
        self['Catch'].add(catch)

    def addRetry(self, retry):
        """Add a Retry to this State"""
        if 'Retry' not in self:
            self['Retry'] = []
        self['Retry'].add(retry)

    def __str__(self):
        return self.Name

class PassState(State):
    def __init__(self, name, **kwargs):
        """
        Args:
            name (string): Name of the state
            kwargs (dict): Arguments passed to State constructor
        """
        super(PassState , self).__init__(name, 'Pass', **kwargs)

class SuccessState(State):
    def __init__(self, name, **kwargs):
        """
        Args:
            name (string): Name of the state
            kwargs (dict): Arguments passed to State constructor
        """
        super(SuccessState, self).__init__(name, 'Succeed', **kwargs)

class FailState(State):
    def __init__(self, name, error, cause, **kwargs):
        """
        Args:
            name (string): Name of the state
            error (string): Failure error name
            cause (string): Failure error cause
            kwargs (dict): Arguments passed to State constructor
        """
        super(FailState, self).__init__(name, 'Fail', **kwargs)
        self['Error'] = error
        self['Cause'] = cause

class TaskState(State):
    def __init__(self, name, resource, **kwargs):
        """
        Args:
            name (string): Name of the state
            resource (string): ARN of AWS resource to invoke
            kwargs (dict): Arguments passed to State constructor
        """
        super(TaskState, self).__init__(name, 'Task', **kwargs)
        self['Resource'] = resource

class WaitState(State):
    def __init__(self, name, seconds=None, timestamp=None, timestamp_path=None, seconds_path=None, **kwargs):
        """
        Args:
            name (string): Name of the state
            seconds (int): Number of seconds to wait
            timestamp (string|Timestamp): Timestamp to wait until
            timestamp_path (string): JsonPath to location of timestamp
            seconds_path (string): JsonPath to location of seconds
            kwargs (dict): Arguments passed to State constructor
        """
        super(WaitState, self).__init__(name, 'Wait', **kwargs)

        if seconds is not None:
            self['Seconds'] = int(seconds)

        if timestamp is not None:
            self['Timestamp'] = str(timestamp)

        if timestamp_path is not None:
            self['TimestampPath'] = str(timestamp_path)

        if seconds_path is not None:
            self['SecondsPath'] = str(seconds_path)

class ParallelState(State):
    def __init__(self, name, branches=None, **kwargs):
        """
        Args:
            name (string): Name of the state
            branches (list): List of List of States to be executed
            kwargs (dict): Arguments passed to State constructor
        """
        super(ParallelState, self).__init__(name, 'Parallel', **kwargs)

        if branches is None:
            branches = []
        self['Branches'] = branches

    def addBranch(self, branch):
        """Add another branch of execution
        
        Args:
            branch (list): List of States to execute"""
        self['Branches'].add(branch)

class ChoiceState(State):
    def __init__(self, name, choices=None, default=None, **kwargs):
        """
        Args:
            name (string): Name of the state
            choices (Choice|[Choice]): Choice(s) that are checked and
                                       potentially executed
            default (string|State): Default state if no Choice matches
            kwargs (dict): Arguments passed to State constructor
        """
        super(ChoiceState, self).__init__(name, 'Choice', **kwargs)

        if choices is None:
            choices = []
        elif type(choices) != list:
            choices = [choices]
        self['Choices'] = choices

        if default is not None:
            self['Default'] = str(default)

    def addChoice(self, choice):
        """Add another Choice to the list of Choices to compare against"""
        self['Choices'].add(choice)

    def setDefault(self, default):
        """Sets the default State to execute if there is no match"""
        self['Default'] = str(default)

class Choice(dict):
    """A ChoiceState Choice wrapping a comparison and reference to state to execute"""

    def __init__(self, variable, op, value, next_ = None):
        """
        Args:
            variable (string): JsonPath of variable to compare
            op (string): AWS Step Function Language operator name
            value (int|float|string|Timestamp): value to compare against
            next (string|State): State to execute if the comparison is True
        """
        super(Choice, self).__init__(Variable = variable)

        self[op] = value
        self.op = op # for __str__ / __repr__

        if next_ is not None:
            self['Next'] = str(next_)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "({} {} {})".format(self['Variable'], self.op, self[self.op])

class NotChoice(dict):
    """Wraper around a Choice that negates the Choice"""

    def __init__(self, value, next_ = None):
        """
        Args:
            value (Choice): Choice to negate
            next (string|State): State to execute if the negated comparison is True
        """
        super(NoteChoice, self).__init__(Not = value)

        if next_ is not None:
            self['Next'] = str(next_)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "(Not {!r})".format(self['Not'])

class AndOrChoice(dict):
    """Wrapper arounds a list of Choices and 'and's or 'or's the results together"""

    def __init__(self, op, values, next_ = None):
        """
        Args:
            op (string): 'and' or 'or' depending on which operator is desired
            values (list): list of two or more Choices to combind the results of
            next (string|State): State to execute if the whole results is True
        """
        super(AndOrChoice, self).__init__()

        if type(values) != list:
            values = [values]

        self[op] = values
        self.op = op # for __str__ / __repr__

        if next_ is not None:
            self['Next'] = str(next_)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        vals = map(repr, self[self.op])
        return "(" + (" {} ".format(self.op.lower())).join(vals) + ")"

class Retry(dict):
    """A retry statement that reruns a state if the given error(s) are encountered"""

    def __init__(self, errors, interval, max, backoff):
        """
        Args:
            errors (string|list): Error(s) to match against
            interval (int): Initial delay (seconds) before retrying state
            max (int): Max number to retries to attempt
            backoff (int|float): Backoff rate applied to interval after each retry
        """
        if type(errors) != list:
            errors = [errors]

        super(Retry, self).__init__(ErrorEquals = errors,
                         IntervalSeconds = int(interval),
                         MaxAttempts = int(max),
                         BackoffRate = float(backoff))

class Catch(dict):
    """A catch statement that executes a state when the given error(s) are encountered"""

    def __init__(self, errors, next, results=None):
        """
        Args:
            errors (string|list): Error(s) to match against
            next (string|State): State to execute if a matched error is encountered
            results (string|None): ResultPath for error information
        """
        if type(errors) != list:
            errors = [errors]

        super(Catch, self).__init__(ErrorEquals = errors,
                         Next = str(next))

        if results:
            self['ResultPath'] = results

def _resolve(actual, defaults):
    """Break the actual arn apart and insert the defaults for the
    unspecified begining parts of the arn (based on position in the
    arn)

    Example:
        actual: 'account_id:function:FUNCTION_NAME'
        defaults: ['arn', 'aws', 'lambda', 'region', 'account_id', 'function']
        return: 'arn:aws:lambda:region:account_id:function:FUNCTION_NAME'
    
    Args:
        actual (string): ARN style string, potentially missing part of the
                         begining of the ARN. Must include the final name of
                         the ARN function
        defaults (list): List of ARN components to use to fill in the missing
                         parts of the actual ARN

    Returns:
        (string): Complete ARN
    """
    actual_ = actual.split(':')
    name = actual_.pop()
    offset = len(defaults) - len(actual_)
    try:
        # Wrap the join because an error should only be produced if we try
        # to use a None value. None can be passed in the defaults if that
        # default value is never used
        vals = defaults[:offset]
        vals.extend(actual_)
        vals.append(name)
        arn = ":".join(vals)
        return arn
    except TypeError:
        raise Exception("One or more of the default values for ARN '{}' was not specified".format(actual))


def Lambda(name, region=None, account=None):
    """Resolve a partial Lambda ARN into a full ARN with the given region/account
    
    Args:
        name (string): Partial ARN
        region (string): AWS region of the Lambda function
        account (string): AWS account id owning the Labmda function
    """

    defaults = ['arn', 'aws', 'lambda', region, account, 'function']
    return _resolve(name, defaults)

def Activity(name, region=None, account=None):
    """Resolve a partial Activity ARN into a full ARN with the given region/account
    
    Args:
        name (string): Partial ARN
        region (string): AWS region of the Activity ARN
        account (string): AWS account id owning the Activity ARN
    """

    defaults = ['arn', 'aws', 'states', region, account, 'activity']
    return _resolve(name, defaults)

class Timestamp(object):
    """Wrapper around a timestamp string.

    Used to determine if a string is in a valid timestamp format and type it
    for the parser
    """

    def __init__(self, timestamp):
        """
        Args:
            timestamp (string): Timestamp string

        Exceptions:
            An exception is thrown if the string is not a valid timestamp
        """
        iso8601.parse_date(timestamp)
        self.timestamp = timestamp

    def __str__(self):
        return self.timestamp




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, unicode_literals

import sys
import time
import json
from datetime import datetime

from botocore.exceptions import ClientError

from .lexer import tokenize_source
from .parser import parse
from .exceptions import StepFunctionError
from .utils import create_session, read


def compile(source, region=None, account_id=None, translate=None, file=sys.stderr, **kwargs):
    """Compile a source step function dsl file into the AWS state machine definition

    Args:
        source (string|Path|file object): Source of step function dsl, passed to read()
        region (string): AWS Region for Lambda / Activity ARNs that need to be filled in
        account_id (string): AWS Account ID for Lambda / Activity ARNs that need to be filled in
        translate (None|function): Function that translates a Lambda / Activity name before
                                   the ARN is completed
        file (file object): Where any error messages are printed (default stderr)
        kwargs (dict): Arguments to be passed to json.dumps() when creating the definition

    Returns:
        string: State machine definition
    """
    try:
        with read(source) as fh:
            if hasattr(fh, 'name'):
                source_name = fh.name
            else:
                source_name = "<unknown>"
            tokens = tokenize_source(fh.readline)

        if translate is None:
            translate = lambda x: x

        machine = parse(tokens, region, account_id, translate)
        def_ = machine.definition(**kwargs)
        return def_
    except StepFunctionError as e:
        print('File "{}", line {}'.format(source_name, e.lineno), file=file)
        print('{}'.format(e.line), file=file)
        print((' ' * e.pos) + '^', file=file)
        print('Syntax Error: {}'.format(str(e)), file=file)
        return None
    #except Exception as e:
    #    print("Unhandled Error: {}".format(e), file=file)

class StateMachine(object):
    """Class for working with and executing AWS Step Function State Machines"""

    def __init__(self, name, **kwargs):
        """
        Args:
            name (string): Name of the state machine
            kwargs (dict): Same arguments as create_session()
        """
        self.name = name
        self.arn = None
        self.session, self.account_id = create_session(**kwargs)
        self.client = self.session.client('stepfunctions')

        resp = self.client.list_state_machines()
        for machine in resp['stateMachines']:
            if machine['name'] == name:
                self.arn = machine['stateMachineArn']
                break

    def _translate(self, function):
        """Default implementation of a function to translate Lambda/Activity names
        before ARNs are created

        Args:
            function (string): Name of Lambda / Activity

        Returns:
            string: Name of the Lambda / Activity
        """
        return function

    def build(self, source, **kwargs):
        """Build the state machine definition from a source (file)

        Region and account id are determined from constructor arguments

        Args:
            source (string|file path|file object): Source of step function dsl
            kwargs (dict): Arguments to be passed to json.dumps() when creating the definition

        Returns:
            string: State machine definition
        """
        region = self.session.region_name
        return compile(source, region, self.account_id, self._translate, **kwargs)

    def _resolve_role(self, role):
        role = role.strip()
        if not role.lower().startswith("arn:aws:iam"):
            client = self.session.client('iam')
            try:
                response = client.get_role(RoleName=role)
                role = response['Role']['Arn']
            except:
                raise Exception("Could not lookup role '{}'".format(role))

        return role

    def create(self, source, role):
        """Create the state machine in AWS from the give source

        If a state machine with the given name already exists an exception is thrown

        Args:
            source (string|file path|file object): Source of step function dsl
            role (string): AWS IAM role for the state machine to execute under
        """
        if self.arn is not None:
            raise Exception("State Machine {} already exists".format(self.arn))

        role = self._resolve_role(role)
        definition = self.build(source)
        # DP TODO: figure out error handling
        #          should allow / return error output
        #          and or throw a custom exception

        resp = self.client.create_state_machine(name = self.name,
                                                definition = definition,
                                                roleArn = role)

        self.arn = resp['stateMachineArn']

    def delete(self, exception=False):
        """Delete the state machine from AWS

        Args:
            exception (boolean): If an excpetion should be thrown if the machine doesn't exist (default: False)
        """
        if self.arn is None:
            if exception:
                raise Exception("State Machine {} doesn't exist yet".format(self.name))
        else:
            resp = self.client.delete_state_machine(stateMachineArn = self.arn)
            self.arn = None

    def start(self, input_, name=None):
        """Start executing the state machine

        If the state machine doesn't exists an exception is thrown

        Args:
            input_ (Json): Json input data for the first state to process
            name (string|None): Name of the execution (default: Name of the state machine)

        Returns:
            string: ARN of the state machine execution, used to get status and output data
        """
        if self.arn is None:
            raise Exception("State Machine {} doesn't exist yet".format(self.name))

        input_ = json.dumps(input_)

        if name is None:
            name = self.name + "-" + datetime.now().strftime("%Y%m%d%H%M%s%f")

        resp = self.client.start_execution(stateMachineArn = self.arn,
                                           name = name,
                                           input = input_)

        arn = resp['executionArn']
        return arn # DP NOTE: Could store ARN in internal dict and return execution name

    def stop(self, arn, error, cause):
        """Stop an execution of the state machine

        Args:
            arn (string): ARN of the execution to stop
            error (string): Error for the stop
            cause (string): Error cause for the stop
        """
        resp = self.client.stop_execution(executionArn = arn,
                                          error = error,
                                          cause = cause)

    def status(self, arn):
        """Get the status of an execution

        Args:
            arn (string): ARN of the execution to get the status of

        Returns:
            string: One of 'RUNNING', 'SUCCEEDED', 'FAILED', 'TIMED_OUT', 'ABORTED'
        """
        resp = self.client.describe_execution(executionArn = arn)
        return resp['status']

    def wait(self, arn, period=10):
        """Wait for an execution to finish and get the results

        Args:
            arn (string): ARN of the execution to get the status of
            period (int): Number of seconds to sleep between polls for status

        Returns:
            dict|None: Dict of Json data or None if there was an error
        """
        while True:
            resp = self.client.describe_execution(executionArn = arn)
            if resp['status'] != 'RUNNING':
                if 'output' in resp:
                    return json.loads(resp['output'])
                else:
                    resp = self.client.get_execution_history(executionArn = arn,
                                                             reverseOrder = True)
                    event = resp['events'][0]
                    for key in ['Failed', 'Aborted', 'TimedOut']:
                        key = 'execution{}EventDetails'.format(key)
                        if key in event:
                            return event[key]
                    raise Exception("Could not locate error output for execution '{}'".format(arn))
            else:
                time.sleep(period)

    def running_arns(self):
        """Query for the ARNs of running executions

        Returns:
            list: List of strings containing the ARNs of all running executions
        """
        resp = self.client.list_executions(stateMachineArn = self.arn,
                                           statusFilter = 'RUNNING')
        return [ex['executionArn'] for ex in resp['executions']]




#!/usr/bin/env python
# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = '0.8'

import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))
def read(filename):
    with open(os.path.join(here, filename), 'r') as fh:
        return fh.read()

def test_suite():
    import unittest
    loader = unittest.TestLoader()
    suite = loader.discover('heaviside', pattern='test_*.py')
    return suite

setup(
    name='heaviside',
    version=__version__,
    packages=['heaviside'],
    url='https://github.com/jhuapl-boss/heaviside',
    license="Apache 2.0 Software License",
    author='Derek Pryor',
    author_email='Derek.Pryor@jhuapl.edu',
    description='Python library and DSL for working with AWS StepFunctions',
    long_description=read('README.md'),
    install_requires=read('requirements.txt').split('\n'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 2.7',
    ],
    keywords=[
        'boss',
        'microns',
        'aws',
        'stepfunctions',
        'dsl',
    ],
    scripts=[
        'bin/heaviside'
    ],
    test_suite='setup.test_suite'
)
